#!/usr/bin/env python
# original code from https://gist.github.com/pwithnall/7bc5f320b3bdf418265a

"""
Gets the current version number.
If in a git repository, it is the current git tag.
Otherwise it is the one contained in the __init__ file.

To use this script, simply import it in your setup.py file
and use the results of get_version() as your package version:

    from version import *

    setup(
        ...
        version=get_version(),
        ...
    )
"""

import codecs
import os
import subprocess


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def version_from_init(rel_path="__init__.py"):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")


def get_version():
    # get git root directory

    # check if "CI" is set, if so, use the version from __init__.py
    if os.environ.get("CI") == "true":
        print("CI environment detected, using __init__.py")
        return version_from_init()

    try:
        cmd = "git describe --tags --match v[0-9]*".split()
        version = subprocess.check_output(cmd).decode().strip()

        # PEP 386 compatibility
        if "-" in version:
            version = ".post".join(version.split("-")[:2])

        # Don't declare a version "dirty" merely because a time stamp has
        # changed. If it is dirty, append a ".dev1" suffix to indicate a
        # development revision after the release.
        with open(os.devnull, "w", encoding="utf8") as fd_devnull:
            subprocess.call(["git", "status"], stdout=fd_devnull, stderr=fd_devnull)

        cmd = "git diff-index --name-only HEAD".split()

        dirty = subprocess.check_output(cmd).decode().strip()

        if dirty != "":
            version += ".dev1"

    except Exception as e:
        # Not a git repository or something went wrong with git
        print(f"{e}\nUnable to get version number from git tags, using __init__.py")
        version = version_from_init()

    return version


if __name__ == "__main__":
    print(get_version())



#!/usr/bin/env python3
"""
 support functions

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
from pathlib import Path

import can
import cantools


def get_axis_id(msg: can.Message) -> int:
    """get axis id from message"""
    return msg.arbitration_id >> 5


def extract_ids(can_id: int) -> tuple[int, int]:
    """get axis_id and cmd_id from can_id"""
    cmd_id = can_id & 0x1F  # Extract lower 5 bits for cmd_id
    axis_id = can_id >> 5  # Shift right by 5 bits to get axis_id
    return axis_id, cmd_id


# pylint: disable=import-outside-toplevel
def get_dbc(name: str = "odrive-cansimple-0.5.6"):
    """get the cantools database"""

    # get relative path to db file
    dbc_path = Path(__file__).parent / f"dbc/{name}.dbc"

    return cantools.database.load_file(dbc_path.as_posix())



# copied from odrive 0.6.7
# NOTE: This file is dangerous because the enums could potentially change between API versions. Should transmit as part of the JSON.
# To regenerate this file, nagivate to the top level of the ODrive repository and run:
#   python Firmware/interface_generator_stub.py --definitions Firmware/odrive-interface.yaml --template tools/enums_template.j2 --output tools/odrive/enums.py

# type: ignore

import enum


class GpioMode(enum.IntEnum):
    DIGITAL = 0
    DIGITAL_PULL_UP = 1
    DIGITAL_PULL_DOWN = 2
    ANALOG_IN = 3
    UART_A = 4
    UART_B = 5
    UART_C = 6
    CAN_A = 7
    I2C_A = 8
    SPI_A = 9
    PWM = 10
    ENC0 = 11
    ENC1 = 12
    ENC2 = 13
    MECH_BRAKE = 14
    STATUS = 15
    BRAKE_RES = 16
    AUTO = 17


class StreamProtocolType(enum.IntEnum):
    FIBRE = 0
    ASCII = 1
    STDOUT = 2
    ASCII_AND_STDOUT = 3
    OTHER = 4


class Protocol(enum.IntFlag):
    SIMPLE = 0x00000001


class AxisState(enum.IntEnum):
    UNDEFINED = 0
    IDLE = 1
    STARTUP_SEQUENCE = 2
    FULL_CALIBRATION_SEQUENCE = 3
    MOTOR_CALIBRATION = 4
    ENCODER_INDEX_SEARCH = 6
    ENCODER_OFFSET_CALIBRATION = 7
    CLOSED_LOOP_CONTROL = 8
    LOCKIN_SPIN = 9
    ENCODER_DIR_FIND = 10
    HOMING = 11
    ENCODER_HALL_POLARITY_CALIBRATION = 12
    ENCODER_HALL_PHASE_CALIBRATION = 13
    ANTICOGGING_CALIBRATION = 14


class ControlMode(enum.IntEnum):
    VOLTAGE_CONTROL = 0
    TORQUE_CONTROL = 1
    VELOCITY_CONTROL = 2
    POSITION_CONTROL = 3


class ComponentStatus(enum.IntEnum):
    NOMINAL = 0
    NO_RESPONSE = 1
    INVALID_RESPONSE_LENGTH = 2
    PARITY_MISMATCH = 3
    ILLEGAL_HALL_STATE = 4
    POLARITY_NOT_CALIBRATED = 5
    PHASES_NOT_CALIBRATED = 6
    NUMERICAL_ERROR = 7
    MISSING_INPUT = 8
    RELATIVE_MODE = 9
    UNCONFIGURED = 10
    OVERSPEED = 11
    INDEX_NOT_FOUND = 12
    BAD_CONFIG = 13
    NOT_ENABLED = 14
    SPINOUT_DETECTED = 15


class ODriveError(enum.IntFlag):
    NONE = 0x00000000
    INITIALIZING = 0x00000001
    SYSTEM_LEVEL = 0x00000002
    TIMING_ERROR = 0x00000004
    MISSING_ESTIMATE = 0x00000008
    BAD_CONFIG = 0x00000010
    DRV_FAULT = 0x00000020
    MISSING_INPUT = 0x00000040
    DC_BUS_OVER_VOLTAGE = 0x00000100
    DC_BUS_UNDER_VOLTAGE = 0x00000200
    DC_BUS_OVER_CURRENT = 0x00000400
    DC_BUS_OVER_REGEN_CURRENT = 0x00000800
    CURRENT_LIMIT_VIOLATION = 0x00001000
    MOTOR_OVER_TEMP = 0x00002000
    INVERTER_OVER_TEMP = 0x00004000
    VELOCITY_LIMIT_VIOLATION = 0x00008000
    POSITION_LIMIT_VIOLATION = 0x00010000
    WATCHDOG_TIMER_EXPIRED = 0x01000000
    ESTOP_REQUESTED = 0x02000000
    SPINOUT_DETECTED = 0x04000000
    BRAKE_RESISTOR_DISARMED = 0x08000000
    THERMISTOR_DISCONNECTED = 0x10000000
    CALIBRATION_ERROR = 0x40000000


class ProcedureResult(enum.IntEnum):
    SUCCESS = 0
    BUSY = 1
    CANCELLED = 2
    DISARMED = 3
    NO_RESPONSE = 4
    POLE_PAIR_CPR_MISMATCH = 5
    PHASE_RESISTANCE_OUT_OF_RANGE = 6
    PHASE_INDUCTANCE_OUT_OF_RANGE = 7
    UNBALANCED_PHASES = 8
    INVALID_MOTOR_TYPE = 9
    ILLEGAL_HALL_STATE = 10
    TIMEOUT = 11
    HOMING_WITHOUT_ENDSTOP = 12
    INVALID_STATE = 13
    NOT_CALIBRATED = 14
    NOT_CONVERGING = 15


class EncoderId(enum.IntEnum):
    NONE = 0
    INC_ENCODER0 = 1
    INC_ENCODER1 = 2
    INC_ENCODER2 = 3
    SENSORLESS_ESTIMATOR = 4
    SPI_ENCODER0 = 5
    SPI_ENCODER1 = 6
    SPI_ENCODER2 = 7
    HALL_ENCODER0 = 8
    HALL_ENCODER1 = 9
    AMT21_ENCODER0 = 10
    AMT21_ENCODER1 = 11
    AMT21_ENCODER2 = 12
    ONBOARD_ENCODER0 = 13
    ONBOARD_ENCODER1 = 14


class SpiEncoderMode(enum.IntEnum):
    DISABLED = 0
    RLS = 1
    AMS = 2
    CUI = 3
    AEAT = 4
    MA732 = 5
    TLE = 6


class InputMode(enum.IntEnum):
    INACTIVE = 0
    PASSTHROUGH = 1
    VEL_RAMP = 2
    POS_FILTER = 3
    MIX_CHANNELS = 4
    TRAP_TRAJ = 5
    TORQUE_RAMP = 6
    MIRROR = 7
    TUNING = 8


class MotorType(enum.IntEnum):
    HIGH_CURRENT = 0
    GIMBAL = 2
    ACIM = 3


class CanError(enum.IntFlag):
    NONE = 0x00000000
    DUPLICATE_CAN_IDS = 0x00000001


def _extend_enum(t, new_vals):
    all = {e.name: e.value for e in t}
    all.update(new_vals)
    if issubclass(t, enum.IntFlag):
        return enum.IntFlag(t.__name__, all)
    else:
        return enum.IntEnum(t.__name__, all)


### Legacy enums (not autogenerated) ###

# Removed during 0.6.0 => 0.6.1


class LegacyODriveError(enum.IntFlag):
    NONE = 0x00000000
    CONTROL_ITERATION_MISSED = 0x00000001
    DC_BUS_UNDER_VOLTAGE = 0x00000002
    DC_BUS_OVER_VOLTAGE = 0x00000004
    DC_BUS_OVER_REGEN_CURRENT = 0x00000008
    DC_BUS_OVER_CURRENT = 0x00000010
    BRAKE_DEADTIME_VIOLATION = 0x00000020
    BRAKE_DUTY_CYCLE_NAN = 0x00000040
    INVALID_BRAKE_RESISTANCE = 0x00000080


class AxisError(enum.IntFlag):
    NONE = 0x00000000
    INVALID_STATE = 0x00000001
    WATCHDOG_TIMER_EXPIRED = 0x00000800
    MIN_ENDSTOP_PRESSED = 0x00001000
    MAX_ENDSTOP_PRESSED = 0x00002000
    ESTOP_REQUESTED = 0x00004000
    HOMING_WITHOUT_ENDSTOP = 0x00020000
    OVER_TEMP = 0x00040000
    UNKNOWN_POSITION = 0x00080000


class MotorError(enum.IntFlag):
    NONE = 0x00000000
    DRV_FAULT = 0x00000008
    CONTROL_DEADLINE_MISSED = 0x00000010
    MODULATION_MAGNITUDE = 0x00000080
    CURRENT_SENSE_SATURATION = 0x00000400
    CURRENT_LIMIT_VIOLATION = 0x00001000
    MODULATION_IS_NAN = 0x00010000
    MOTOR_THERMISTOR_OVER_TEMP = 0x00020000
    FET_THERMISTOR_OVER_TEMP = 0x00040000
    TIMER_UPDATE_MISSED = 0x00080000
    CURRENT_MEASUREMENT_UNAVAILABLE = 0x00100000
    CONTROLLER_FAILED = 0x00200000
    I_BUS_OUT_OF_RANGE = 0x00400000
    BRAKE_RESISTOR_DISARMED = 0x00800000
    SYSTEM_LEVEL = 0x01000000
    BAD_TIMING = 0x02000000
    UNKNOWN_PHASE_ESTIMATE = 0x04000000
    UNKNOWN_PHASE_VEL = 0x08000000
    UNKNOWN_TORQUE = 0x10000000
    UNKNOWN_CURRENT_COMMAND = 0x20000000
    UNKNOWN_CURRENT_MEASUREMENT = 0x40000000
    UNKNOWN_VBUS_VOLTAGE = 0x80000000
    UNKNOWN_VOLTAGE_COMMAND = 0x100000000
    UNKNOWN_GAINS = 0x200000000
    CONTROLLER_INITIALIZING = 0x400000000


class SensorlessEstimatorError(enum.IntFlag):
    NONE = 0x00000000
    UNSTABLE_GAIN = 0x00000001
    UNKNOWN_CURRENT_MEASUREMENT = 0x00000002


# Removed during 0.5.4 => 0.6.0:


class EncoderMode(enum.IntEnum):
    INCREMENTAL = 0
    HALL = 1
    SINCOS = 2
    SPI_ABS_CUI = 256
    SPI_ABS_AMS = 257
    SPI_ABS_AEAT = 258
    SPI_ABS_RLS = 259
    SPI_ABS_MA732 = 260


MotorError = _extend_enum(
    MotorError,
    {
        "PHASE_RESISTANCE_OUT_OF_RANGE": 0x00000001,
        "PHASE_INDUCTANCE_OUT_OF_RANGE": 0x00000002,
        "UNBALANCED_PHASES": 0x800000000,
    },
)


class ControllerError(enum.IntFlag):
    NONE = 0x00000000
    OVERSPEED = 0x00000001
    INVALID_INPUT_MODE = 0x00000002
    UNSTABLE_GAIN = 0x00000004
    INVALID_MIRROR_AXIS = 0x00000008
    INVALID_LOAD_ENCODER = 0x00000010
    INVALID_ESTIMATE = 0x00000020
    INVALID_CIRCULAR_RANGE = 0x00000040
    SPINOUT_DETECTED = 0x00000080


class EncoderError(enum.IntFlag):
    NONE = 0x00000000
    UNSTABLE_GAIN = 0x00000001
    CPR_POLEPAIRS_MISMATCH = 0x00000002
    NO_RESPONSE = 0x00000004
    UNSUPPORTED_ENCODER_MODE = 0x00000008
    ILLEGAL_HALL_STATE = 0x00000010
    INDEX_NOT_FOUND_YET = 0x00000020
    ABS_SPI_TIMEOUT = 0x00000040
    ABS_SPI_COM_FAIL = 0x00000080
    ABS_SPI_NOT_READY = 0x00000100
    HALL_NOT_CALIBRATED_YET = 0x00000200


# Enums in 0.5.2, 0.5.3 and 0.5.4 are identical

# Removed during 0.5.1 => 0.5.2:

AxisState = _extend_enum(AxisState, {"SENSORLESS_CONTROL": 5})


class ThermistorCurrentLimiterError(enum.IntFlag):
    NONE = 0x00000000
    OVER_TEMP = 0x00000001


AxisError = _extend_enum(
    AxisError,
    {
        "DC_BUS_UNDER_VOLTAGE": 0x00000002,
        "DC_BUS_OVER_VOLTAGE": 0x00000004,
        "CURRENT_MEASUREMENT_TIMEOUT": 0x00000008,
        "BRAKE_RESISTOR_DISARMED": 0x00000010,
        "MOTOR_DISARMED": 0x00000020,
        "MOTOR_FAILED": 0x00000040,
        "SENSORLESS_ESTIMATOR_FAILED": 0x00000080,
        "ENCODER_FAILED": 0x00000100,
        "CONTROLLER_FAILED": 0x00000200,
        "POS_CTRL_DURING_SENSORLESS": 0x00000400,
    },
)


class LockinState(enum.IntEnum):
    INACTIVE = 0
    RAMP = 1
    ACCELERATE = 2
    CONST_VEL = 3


MotorError = _extend_enum(
    MotorError,
    {
        "ADC_FAILED": 0x00000004,
        "NOT_IMPLEMENTED_MOTOR_TYPE": 0x00000020,
        "BRAKE_CURRENT_OUT_OF_RANGE": 0x00000040,
        "BRAKE_DEADTIME_VIOLATION": 0x00000100,
        "UNEXPECTED_TIMER_CALLBACK": 0x00000200,
        "BRAKE_DUTY_CYCLE_NAN": 0x00002000,
        "DC_BUS_OVER_REGEN_CURRENT": 0x00004000,
        "DC_BUS_OVER_CURRENT": 0x00008000,
    },
)


class ArmedState(enum.IntEnum):
    DISARMED = 0
    WAITING_FOR_TIMINGS = 1
    WAITING_FOR_UPDATE = 2
    ARMED = 3


class DrvFault(enum.IntFlag):
    NO_FAULT = 0x00000000
    FET_LOW_C_OVERCURRENT = 0x00000001
    FET_HIGH_C_OVERCURRENT = 0x00000002
    FET_LOW_B_OVERCURRENT = 0x00000004
    FET_HIGH_B_OVERCURRENT = 0x00000008
    FET_LOW_A_OVERCURRENT = 0x00000010
    FET_HIGH_A_OVERCURRENT = 0x00000020
    OVERTEMPERATURE_WARNING = 0x00000040
    OVERTEMPERATURE_SHUTDOWN = 0x00000080
    P_VDD_UNDERVOLTAGE = 0x00000100
    G_VDD_UNDERVOLTAGE = 0x00000200
    G_VDD_OVERVOLTAGE = 0x00000400


# Changed during 0.5.1 => 0.5.2:
# PROTOCOL_SIMPLE                          = 0
# changed to
# PROTOCOL_SIMPLE                          = 0x00000001

# MotorError.ERROR_CURRENT_UNSTABLE renamed to MotorError.CURRENT_LIMIT_VIOLATION


def _load_legacy_names():
    """
    Loads legacy names for all enumerators.
    e.g. AxisState.IDLE => AXIS_STATE_IDLE

    This was introduced in odrive package version 0.6.0.

    This will be removed in future versions of the odrive package and is only
    included for a transition period until we have updated the legacy
    documentation.
    """
    import sys

    module = sys.modules[__name__]

    def to_macro_case(s):
        return "".join([("_" + c) if c.isupper() else (c.upper()) for c in s]).lstrip(
            "_"
        )

    legacy_enum_names = [
        "StreamProtocolType",
        "Protocol",
        "AxisState",
        "ControlMode",
        "ComponentStatus",
        "ProcedureResult",
        "EncoderId",
        "SpiEncoderMode",
        "InputMode",
        "MotorType",
        "LegacyODriveError",
        "CanError",
        "AxisError",
        "MotorError",
        "SensorlessEstimatorError",
        "GpioMode",
        "EncoderError",
        "ControllerError",
        "ThermistorCurrentLimiterError",
        "LockinState",
        "EncoderMode",
        "ArmedState",
        "DrvFault",
    ]
    for var in legacy_enum_names:
        t = getattr(module, var)
        for item in t:
            setattr(
                module,
                (
                    "LEGACY_ODRIVE_ERROR"
                    if var == "LegacyODriveError"
                    else to_macro_case(var)
                )
                + "_"
                + item.name,
                item.value,
            )


_load_legacy_names()



#!/usr/bin/env python3

# interface generated by gen_dbc_interface.py
# post-edited manually.

import asyncio
from typing import Optional
from abc import ABC, abstractmethod


class DbcInterface(ABC):
    """translation of DBC file to python interface"""

    def __init__(self) -> None:
        self.setpoint: float = 0.0  # keep track of setpoint

    @property
    @abstractmethod
    def axis_state(self) -> str:
        pass

    @abstractmethod
    async def _request(self, msg_name: str, timeout: float = 0.5) -> dict:
        pass

    @abstractmethod
    def _send_message(
        self, msg_name: str, msg_dict: Optional[dict] = None, rtr: bool = False
    ):
        pass

    # -------manage axis
    def clear_errors(self):
        self._send_message("Clear_Errors")

    def reboot(self):
        self._send_message("Reboot")

    # -------getters

    async def get_motor_error(self) -> dict:
        return await self._request("Get_Motor_Error")

    async def get_encoder_error(self) -> dict:
        return await self._request("Get_Encoder_Error")

    async def get_sensorless_error(self) -> dict:
        return await self._request("Get_Sensorless_Error")

    async def get_encoder_estimates(self) -> dict:
        return await self._request("Get_Encoder_Estimates")

    async def get_encoder_count(self) -> dict:
        return await self._request("Get_Encoder_Count")

    async def get_iq(self) -> dict:
        return await self._request("Get_Iq")

    async def get_sensorless_estimates(self) -> dict:
        return await self._request("Get_Sensorless_Estimates")

    async def get_bus_voltage_current(self) -> dict:
        return await self._request("Get_Bus_Voltage_Current")

    async def get_adc_voltage(self) -> dict:
        return await self._request("Get_ADC_Voltage")

    async def get_controller_error(self) -> dict:
        return await self._request("Get_Controller_Error")

    # -------setters

    def set_axis_node_id(self, axis_node_id: int):
        self._send_message("Set_Axis_Node_ID", {"Axis_Node_ID": axis_node_id})

    async def set_axis_state(self, axis_requested_state: str | int):
        """request axis state and wait for it to be reached"""
        self._send_message(
            "Set_Axis_State", {"Axis_Requested_State": axis_requested_state}
        )
        while self.axis_state != axis_requested_state:
            await asyncio.sleep(0.1)

    def set_controller_mode(self, control_mode: str | int, input_mode: str | int):
        self._send_message(
            "Set_Controller_Mode",
            {"Control_Mode": control_mode, "Input_Mode": input_mode},
        )

    def set_input_pos(
        self, input_pos: float, vel_ff: float = 0.0, torque_ff: float = 0.0
    ):
        self._send_message(
            "Set_Input_Pos",
            {"Input_Pos": input_pos, "Vel_FF": vel_ff, "Torque_FF": torque_ff},
        )
        self.setpoint = input_pos

    def set_input_vel(self, input_vel: float, input_torque_ff: float = 0.0):
        self._send_message(
            "Set_Input_Vel",
            {"Input_Vel": input_vel, "Input_Torque_FF": input_torque_ff},
        )
        self.setpoint = input_vel

    def set_input_torque(self, input_torque: float):
        self._send_message("Set_Input_Torque", {"Input_Torque": input_torque})
        self.setpoint = input_torque

    def set_limits(self, velocity_limit: float, current_limit: float):
        self._send_message(
            "Set_Limits",
            {"Velocity_Limit": velocity_limit, "Current_Limit": current_limit},
        )

    def set_traj_vel_limit(self, traj_vel_limit: float):
        self._send_message("Set_Traj_Vel_Limit", {"Traj_Vel_Limit": traj_vel_limit})

    def set_traj_accel_limits(self, traj_accel_limit: float, traj_decel_limit: float):
        self._send_message(
            "Set_Traj_Accel_Limits",
            {
                "Traj_Accel_Limit": traj_accel_limit,
                "Traj_Decel_Limit": traj_decel_limit,
            },
        )

    def set_traj_inertia(self, traj_inertia: float):
        self._send_message("Set_Traj_Inertia", {"Traj_Inertia": traj_inertia})

    def set_linear_count(self, position: int):
        self._send_message("Set_Linear_Count", {"Position": position})

    def set_pos_gain(self, pos_gain: float):
        self._send_message("Set_Pos_Gain", {"Pos_Gain": pos_gain})

    def set_vel_gains(self, vel_gain: float, vel_integrator_gain: float):
        self._send_message(
            "Set_Vel_Gains",
            {"Vel_Gain": vel_gain, "Vel_Integrator_Gain": vel_integrator_gain},
        )



#!/usr/bin/env python3
"""
 ODrive CAN driver

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging
import threading
from enum import Enum
from typing import Callable, Optional

import can

from odrive_can.interface import DbcInterface
from odrive_can.timer import Timer
from odrive_can.utils import extract_ids, get_axis_id, get_dbc


# message timeout in seconds
MESSAGE_TIMEOUT = 1.0  # message expiration time & can timeout
CUSTOM_TIMEOUTS = {"Heartbeat": 0.5}


dbc = get_dbc()


class DriveError(Exception):
    """ODrive drive error"""


class HeartbeatError(Exception):
    """No heartbeat error"""


class CommandId(Enum):
    """short list of command IDs, used for ignoring messages"""

    HEARTBEAT = 0x01
    ENCODER_ESTIMATE = 0x09


class CanMsg:
    """class to manage CAN messages"""

    def __init__(self, msg: can.Message):
        db_msg = dbc.get_message_by_frame_id(msg.arbitration_id)
        self.name = db_msg.name.split("_", 1)[1]  # remove "AxisX_" prefix
        self.data = db_msg.decode(msg.data)

        self.axis_id = get_axis_id(msg)

        timeout = CUSTOM_TIMEOUTS.get(self.name, MESSAGE_TIMEOUT)

        self._timer = Timer(timeout=timeout)

    def is_expired(self):
        """check if timer has expired"""
        return self._timer.is_timeout()

    def __str__(self):
        return f"{self.name}: {self.data}"

    def __repr__(self):
        return str(self)


class ODriveCAN(DbcInterface):
    """odrive CAN driver"""

    def __init__(
        self,
        axis_id: int = 0,
        interface: str = "can0",
        interface_type: str = "socketcan",
    ):
        super().__init__()
        self._log = logging.getLogger(f"odrive.{axis_id}")
        self._axis_id = axis_id

        self._bus = can.interface.Bus(channel=interface, interface=interface_type)

        # tried to use notifier, but it does not work well with asyncio
        # did not succeed in using can.AsyncBufferedReader either, latest state can be found
        # on `remove-trhead` branch. For now, use a separate thread to read messages
        # self._notifier = can.Notifier(self._bus, [self._message_handler])

        self._recieve_thread: Optional[threading.Thread] = None
        self._msg_queue: asyncio.Queue = asyncio.Queue()

        self._last_heartbeat: Optional[CanMsg] = None

        self._ignored_messages: set = set()  # message ids to ignore

        self._response_queue: asyncio.Queue = asyncio.Queue(1)  # response queue
        self._request_id: int = 0  # set to msg.arbitration_id by _send_message

        # called on incoming position message
        self.feedback_callback: Optional[Callable[[CanMsg, ODriveCAN], None]] = None

        self._running = True  # flag to stop loops

    def check_alive(self):
        """check if axis is alive, rasie an exception if not"""
        if self._last_heartbeat is None:
            raise HeartbeatError("Error: No heartbeat message received.")

        if self._last_heartbeat.is_expired():
            raise HeartbeatError("Error: Heartbeat message timeout.")

    def check_errors(self):
        """Check if axis is in error and raise an exception if so."""
        if self._last_heartbeat is None:
            raise HeartbeatError("Error: No heartbeat message received.")

        msg = self._last_heartbeat

        if msg.data["Axis_Error"] != "NONE":
            raise DriveError(f"Axis Error: {msg.data['Axis_Error']}")

        for field in [
            "Motor_Error_Flag",
            "Encoder_Error_Flag",
            "Controller_Error_Flag",
        ]:
            if msg.data[field] != 0:
                raise DriveError(f"{field} Error Detected")

    def ignore_message(self, cmd_id: CommandId):
        """ignore message by command ID"""
        self._ignored_messages.add(cmd_id.value)

    def allow_message(self, cmd_id: CommandId):
        """allow message by command ID"""
        self._ignored_messages.remove(cmd_id.value)

    @property
    def axis_state(self) -> str:
        """get axis state"""
        if self._last_heartbeat is None:
            raise HeartbeatError("Error: No heartbeat message received.")

        return self._last_heartbeat.data["Axis_State"]

    async def start(self):
        """start driver"""
        self._log.info(
            f"Starting. axis_id={self._axis_id}, bus={self._bus.channel_info}"
        )

        loop = asyncio.get_running_loop()  # Get the current asyncio event loop
        self._recieve_thread = threading.Thread(
            target=self._can_reader_thread, args=(loop,), daemon=True
        )
        self._recieve_thread.start()

        asyncio.create_task(self._message_handler())

        # wait for first heartbeat
        self._log.info("waiting for first heartbeat")
        while self._last_heartbeat is None:
            await asyncio.sleep(0.1)

        self._log.info("started")

    def stop(self):
        """stop driver"""
        self._log.info("stopping driver")
        self._send_message("Set_Axis_State", {"Axis_Requested_State": "IDLE"})
        self._running = False

        self._recieve_thread.join()

    # ------------------- private -------------------
    async def _request(self, msg_name: str, timeout: float = 0.5) -> dict:
        """Send an RTR message and wait for the response with a timeout."""

        # check if another request is in progress
        if self._request_id != 0:
            raise RuntimeError("another request is already in progress")

        # self._response_event.clear()  # Reset the event before waiting

        # Send the request
        self._send_message(msg_name, rtr=True)

        result = {}
        try:
            # Wait for the response with a timeout
            msg = await asyncio.wait_for(self._response_queue.get(), timeout)
            self._log.debug("awaited response")
            result = msg.data
        except asyncio.TimeoutError as error:
            # Handle the timeout
            self._log.error(f"Timeout waiting for response to {msg_name}")
            raise TimeoutError(f"Timeout waiting for response to {msg_name}") from error

        finally:
            self._request_id = 0  # Reset the request ID

        return result

    def _send_message(
        self, msg_name: str, msg_dict: Optional[dict] = None, rtr: bool = False
    ):
        """send message by name. If no msg_dict is provided, use zeros
        msg_name is the name of the message without the "AxisX_" prefix
        """
        msg = dbc.get_message_by_name(f"Axis{self._axis_id}_{msg_name}")
        if rtr:
            # For RTR messages, don't specify the data field
            msg = can.Message(
                arbitration_id=msg.frame_id,
                is_extended_id=False,
                is_remote_frame=True,
            )
            # set request id for response
            self._request_id = msg.arbitration_id
        else:
            full_msg_dict = {signal.name: 0 for signal in msg.signals}
            if msg_dict is not None:
                full_msg_dict.update(msg_dict)

            data = msg.encode(full_msg_dict)
            msg = can.Message(
                arbitration_id=msg.frame_id,
                data=data,
                is_extended_id=False,
            )
        try:
            self._bus.send(msg)  # type: ignore
        except can.CanError as error:
            self._log.error(error)

    def __del__(self):
        """destructor"""
        if hasattr(self, "_bus"):
            self._bus.shutdown()

    def _can_reader_thread(self, loop):
        """receive can messages, filter and put them into the queue"""
        while self._running:
            try:
                msg = self._bus.recv(MESSAGE_TIMEOUT)
                if not msg:
                    self._log.warning("can timeout")
                    continue

                axis_id, cmd_id = extract_ids(msg.arbitration_id)

                # Ignore messages that aren't for this axis
                if axis_id != self._axis_id:
                    continue

                # Ignore messages that were requested to be ignored, this is used
                # to increase performance especially for frequent encoder updates
                if cmd_id in self._ignored_messages:
                    self._log.debug(f"ignoring {cmd_id=}")
                    continue

                # RTR messages are requests for data, they don't have a data payload
                # no RTR messages are sent by the odrive, this should not happen
                if msg.is_remote_frame:
                    self._log.warning("RTR message received")
                    continue

                self._log.debug(f"< {axis_id=} {cmd_id=}")

                asyncio.run_coroutine_threadsafe(
                    self._msg_queue.put((cmd_id, msg)), loop
                )
            except Exception as e:  # pylint: disable=broad-except
                self._log.error(f"Error in CAN reader thread: {e}")

        self._log.debug("CAN reader thread stopped")

    async def _message_handler(self):
        """handle received message"""

        while self._running:
            cmd_id, msg = await self._msg_queue.get()
            self._msg_queue.task_done()
            try:
                # process message
                can_msg = CanMsg(msg)

                # check if this is a response to a request
                if msg.arbitration_id == self._request_id:
                    self._log.debug("response received")
                    await self._response_queue.put(can_msg)

                # call position callback
                if cmd_id == CommandId.ENCODER_ESTIMATE.value:
                    if self.feedback_callback is not None:
                        self.feedback_callback(can_msg, self)

                # handle heartbeat
                if cmd_id == CommandId.HEARTBEAT.value:
                    self._log.debug(f"heartbeat: {can_msg.data}")
                    self._last_heartbeat = can_msg

            except KeyError:
                # If the message ID is not in the DBC file, print the raw message
                self._log.warning(f"Unkown message: {msg}")

        self._log.debug("message handler stopped")



from odrive_can import get_dbc, extract_ids


def test_get_db():
    """check database loading"""

    _ = get_dbc()


def test_message_ids():
    dbc = get_dbc()
    for axis_id in range(7):
        dbc_msg = dbc.get_message_by_name(f"Axis{axis_id}_Heartbeat")

        # heartbeat message 0x01
        can_id = axis_id << 5 | 0x01
        assert dbc_msg.frame_id == can_id
        ids = extract_ids(can_id)
        assert ids == (axis_id, 0x01)

        # motor error  0x03
        dbc_msg = dbc.get_message_by_name(f"Axis{axis_id}_Get_Motor_Error")
        can_id = axis_id << 5 | 0x03
        assert dbc_msg.frame_id == can_id
        ids = extract_ids(can_id)
        assert ids == (axis_id, 0x03)



#!/usr/bin/env python3
"""
 Demonstration of velocity control using CAN interface

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging

from odrive_can.odrive import ODriveCAN
from odrive_can.setpoints import sawtooth_generator
from odrive_can.tools import UDP_Client

from .position_control import feedback_callback

SETPOINT_DELAY = 0.1


log = logging.getLogger("pos_ctl")
udp = UDP_Client()


async def configure_controller(drv: ODriveCAN):
    """setup control parameters"""

    # reset encoder
    drv.set_linear_count(0)

    drv.set_controller_mode("VELOCITY_CONTROL", "VEL_RAMP")

    # set position control mode
    await drv.set_axis_state("CLOSED_LOOP_CONTROL")
    drv.check_errors()


async def main_loop(drv: ODriveCAN, amplitude: float = 40.0):
    """velocity control demo"""

    log.info("-----------Running velocity control-----------------")

    drv.feedback_callback = feedback_callback
    await drv.start()

    await asyncio.sleep(0.5)
    drv.check_alive()
    drv.clear_errors()
    drv.check_errors()

    await configure_controller(drv)

    # make setpoint generator
    setpoint_gen = sawtooth_generator(roc=10.0, max_val=40.0)

    drv.set_input_pos(amplitude)

    await asyncio.sleep(2)

    try:
        while True:
            drv.check_errors()
            setpoint = next(setpoint_gen)

            drv.set_input_vel(setpoint)
            await asyncio.sleep(SETPOINT_DELAY)

    except KeyboardInterrupt:
        log.info("Stopping")
    finally:
        drv.stop()
        await asyncio.sleep(0.5)


def main(axis_id: int, interface: str, amplitude: float = 40.0):
    print("Starting velocity control demo, press CTRL+C to exit")
    drv = ODriveCAN(axis_id, interface)

    try:
        asyncio.run(main_loop(drv, amplitude))
    except KeyboardInterrupt:
        log.info("KeyboardInterrupt")


if __name__ == "__main__":
    import coloredlogs  # type: ignore

    from odrive_can import LOG_FORMAT, TIME_FORMAT  # pylint: disable=ungrouped-imports

    coloredlogs.install(level="INFO", fmt=LOG_FORMAT, datefmt=TIME_FORMAT)

    main(1, "slcan0")



__version__ = "0.7.0"

from pathlib import Path

from .utils import get_axis_id, get_dbc, extract_ids
from .odrive import ODriveCAN, CanMsg


LOG_FORMAT = "%(asctime)s [%(name)s] %(filename)s:%(lineno)d - %(message)s"
TIME_FORMAT = "%H:%M:%S.%f"



#!/usr/bin/env python3
"""
 Simple linear model for generating setpoints.

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""


from typing import Optional
import math
import time


def sign(x):
    """sign of a number"""
    if x == 0:
        return 0
    return math.copysign(1, x)


class LinearModel:
    """Simple linear model for generating setpoints."""

    __slots__ = ("val", "roc", "setpoint", "min_val", "max_val", "_last_update")

    def __init__(
        self,
        roc: float,
        val: float = 0.0,
        setpoint: Optional[float] = None,
        min_val: Optional[float] = None,
        max_val: Optional[float] = None,
    ):
        """
        Args:
            roc (float): rate of change / sec
            val (float): current value
            setpoint (float, optional): target value.
            max_val (float, optional): maximum value
            min_val (float, optional): minimum value
        """

        self.val = val
        self.roc = roc
        if setpoint is None:
            self.setpoint = val
        else:
            self.setpoint = setpoint

        self.min_val = min_val
        self.max_val = max_val
        self._last_update = time.time()

    def step(self, dt: Optional[float] = None) -> float:
        """perform timestep, return actual dt"""

        if dt is None:
            dt = time.time() - self._last_update
            self._last_update = time.time()

        if dt < 0:
            raise ValueError(f"dt must be positive, got {dt=} ")

        error = self.setpoint - self.val
        step = sign(error) * self.roc * dt

        if abs(step) > abs(error):
            self.val += error
        else:
            self.val += step

        if self.max_val is not None:
            self.val = min(self.val, self.max_val)

        if self.min_val is not None:
            self.val = max(self.val, self.min_val)

        return dt

    def __repr__(self) -> str:
        return f"LinearModel(val={self.val}, setpoint={self.setpoint}, roc={self.roc})"



# type: ignore

"""The setup script."""


import codecs
import os
import os.path

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")


# please keep this lean and mean. Add dev requirements to .devcontainer/requirments.txt
requirements = ["click", "cantools", "python-can", "coloredlogs"]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Jev Kuznetsov",
    author_email="jev.kuznetsov@gmail.com",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.11",
    ],
    description="Use odrive motion controller with CAN inteface",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    install_requires=requirements,
    keywords="",
    name="odrive_can",
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    package_data={
        "odrive_can": ["dbc/*"],
    },
    test_suite="tests",
    tests_require=test_requirements,
    url="",
    version=get_version("src/odrive_can/__init__.py"),
    # zip_safe=False,
    entry_points={"console_scripts": ["odrive_can=odrive_can.cli:cli"]},
)



#!/usr/bin/env python3
"""
 Tools to assist with odrive development

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""
import time
import socket
import json
import logging


class UDP_Client:
    """send data to UDP server, used for plotting with plotjuggler"""

    def __init__(self, host: str = "127.0.0.1", port: int = 5005):
        self._host = host
        self._port = port
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self._log = logging.getLogger(self.__class__.__name__)

    def send(self, data: dict, add_timestamp: bool = True):
        """send data to UDP server"""
        if add_timestamp:
            data["ts"] = time.time()

        try:
            self._sock.sendto(json.dumps(data).encode(), (self._host, self._port))
        except Exception as e:  # pylint: disable=broad-except
            self._log.error(f"Failed to send data: {e}")



#!/usr/bin/env python3
"""
 Handy setpoint generators for testing etc.

 Copyright (c) 2023 ROX Automation - Jev Kuznetsov
"""

import time
from odrive_can.linear_model import LinearModel


def sawtooth_generator(roc: float = 10.0, max_val: float = 20.0):
    model = LinearModel(
        roc=roc, val=0.0, setpoint=100.0, min_val=-max_val, max_val=max_val
    )
    t = time.time()

    while True:
        dt = time.time() - t  # Calculate time elapsed since last update of t

        model.step(dt)
        # print(f"{dt=}, {model.val=} {t=}")
        if model.val >= max_val:
            model.val = -max_val

        t = time.time()
        yield model.val  # Yield the current value
